# -------------------------------------------------------------------------
# Copyright (c) 2020, PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


__path__ = __import__("pkgutil").extend_path(__path__, __name__)
from . import exchange, range, name

CONSUMER_EXCHANGE = 'CONSUMER'
PRODUCER_EXCHANGE = 'PRODUCER'